/* Convert a Multiface 3 snapshot to .SNA, or vice versa.
   Convert a Multiface 3 snapshot to .Z80.

   .SNA format as described by Arnt Gulbrandsen, 2/2/1994 on Usenet.
	 M3 format as deduced by me: <http://sable.ox.ac.uk/~sjoh0132/m3data.html>
   .Z80 format as described by Gerton Lunter in Z80.DOC.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if CPM                 /* CP/M doesn't support argv[0] */
#define AV0 "M3CONV"
#else  /* CPM */
#define AV0 argv[0]
#endif /* CPM */

#ifndef SEEK_SET        /* Hi-tech C doesn't define SEEK_SET */
#define SEEK_SET 0
#endif

typedef unsigned char uchr;  /* for brevity */

uchr snahdr[27];
uchr zhdr[55];
uchr image[49152];
uchr image2[32768];

uchr m3loader[0x0169] = {

  'P','L','U','S','3','D','O','S',0x1A,         /* +3DOS signature */
  1,0,                                          /* +3DOS v1.0      */
  0x69,1,0,0,                                   /* file length     */
  0,                                            /* +3 BASIC        */
  0xE9,0,                                       /* +3 image length */
  1,0,                                          /* autostart line  */
  0xE9,0,                                       /* program length  */
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,/* unused          */
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,/* unused          */
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,/* unused          */
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,/* unused          */
  0,0,0,0,0,0,0,0,0,0,0,0,0,                    /* unused          */
  0xB5,                                         /* header checksum */

  0x00,0x01,0x91,0x00,0xEA,0xF3,0x01,0x82,0x00, /* BASIC loader:   */
  0x2A,0x53,0x5C,0x09,0x4E,0x23,0x46,0x3E,0xC3, /* 1 REM <Z80 m/c> */
  0x32,0x7F,0x5C,0xED,0x43,0x80,0x5C,0x01,0xFD,
  0x7F,0x3E,0x10,0xED,0x79,0x3A,0x67,0x5B,0xF6,
  0x04,0x06,0x1F,0xED,0x79,0x3E,0xFE,0xCD,0x01,
  0x16,0xDB,0x3F,0x3A,0x00,0x00,0xFE,0xC3,0x28,
  0x10,0x01,0x74,0x00,0x2A,0x53,0x5C,0x09,0x06,
  0x0E,0x7E,0xD7,0x23,0x10,0xFB,0x18,0xE7,0xDB,
  0xBF,0x2A,0x53,0x5C,0xE5,0x01,0x74,0x00,0x09,
  0x06,0x07,0x7E,0xD7,0x23,0x10,0xFB,0xE1,0x01,
  0xBC,0x00,0x09,0x06,0x07,0x7E,0xD7,0x23,0x10,
  0xFB,0xDB,0x3F,0xC3,0x7F,0x5C,
  0xFF,' ',' ',' ',' ',' ',' ',' ',' ',         /* datafile filename        */
  0x20,0xA0,                                    /* datafile block 1 length  */
  0x16,0x14,0x0C,0x12,0x01,0x11,0x07,           /* AT 20,12; FLASH 1; INK 7 */
  'N','O','T',' ','O','N','!',
  0x71,0x00,                                    /* disk, not tape           */
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF,      /* 8912 registers (set to   */
  0x00,0x00,0x00,0x00,0x00,0x00,0xFF,0x00,      /* silence)                 */
  0x0D,0x00,                                    /* end of line              */

  0x00,0x02,0x4F,0x00,                          /* line 2             */
  0xE7,0xBB,0xA7,':',                           /* BORDER SQR PI:     */
  0xD9,0xBB,0xA7,':',                           /* INK SQR PI:        */
  0xDA,0xBB,0xA7,':',                           /* PAPER SQR PI:      */
  0xFD,0xB0,'"','2','4','5','4','3','"',':',    /* CLEAR VAL "24543": */
  0xF5,'"',0x16,0x14,0x09,0x12,0x01,0x11,0x07,  /* PRINT "M3 LOADING" */
  'M','3',' ','L','O','A','D','I','N','G','"',':',
  0xF9,0xC0,'(',0xB0,'"','2','5','6','"','*',   /* RAND USR (PROG)+5  */
  0xBE,0xB0,'"','2','3','6','3','6','"','+',
  0xBE,0xB0,'"','2','3','6','3','5','"','+',
  0xB0,'"','5','"',')',
  0x0D                                          /* End of program.    */
  };
uchr m3dhdr[0xD3] = {
  0x00,                                         /* Not compressed     */
  0x03,
  '<','M','3','C','O','N','V','2','>',0xFF,0x00,0x1B,0x00,0x40,
  0xFF,0xFF,0x00,0x00,0xE8,0x5F,0xB3,0x60,0x00,0x00,0xE2,0x98,0x12,0x14,0xC2,
  0xF8,0x00,0x00,0x00,0x00,0x00,0x1C,0xCC,0x11,'*','M','3','C','O','N','V','2',
  0xFF,0x00,0xE9,0x00,0x01,0x00,0xE9,0x00,0x00,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFB,0xD8,
  0x3A,0x5C,0x00,0x00,0x21,0x15,0x9B,0x36,0x38,0x00,0x6D,0x7E,0x01,0x06,0xDA,
  0x21,0x10,0xFE,0x40,0x3F,0x5B,0x97,0x10,0x01,0xA2,0x61,0x00,0xF5,0xDB,0xBF,
  0xF1,0xCD,0x54,0x60,0xF3,0xF5,0xDB,0x3F,0xF1,0xC9,0xDD,0xE9,0x21,0x6B,0x00,
  0xF3,0xF5,0xDB,0x3F,0xF1,0xE9,0xF5,0xDB,0xBF,0xF1,0xCD,0x5E,0x60,0x18,0xE6,
  0xDB,0xBF,0xFD,0xCB,0x01,0xAE,0xFD,0xCB,0x01,0xDE,0xFD,0xCB,0x30,0xDE,0xFB,
  0x76,0xF3,0xFD,0xCB,0x01,0x6E,0x28,0xE9,0x11,0x22,0x00,0x21,0xC8,0x00,0xCD,
  0xB5,0x03,0xF3,0xDB,0x3F,0x3A,0x08,0x5C,0xC9,0xF5,0xDB,0xBF,0xF1,0xD7,0x18,
  0xB8,0xDB,0xBF,0xED,0xB0,0x18,0xB2,0xDB,0xBF,0xCD,0xE5,0x19,0x18,0xAB,0xDB,
  0xBF,0xCD,0x55,0x16,0x18,0xA4,0xDB,0xBF,0x4E,0x18,0x9F,0x2A,0x2E,0x2A,0xFF
  };
char border=0;
char lock=0;
char smmap[25]= { 0x5f, 0x54, 0x55, 0x52, 0x53, 0x50, 0x51, 0x56,
                  0x57, 0x60, 0x61, 0x5a, 0x5b, 0x58, 0x59, 0x4c,
                  0x4d, 0x4e, 0x4f, 0x5e, 0x5d, 0x62, 0x63, 0x64,
                  0x65 };  /* M3 to .SNA register mapping:     *
                            * snahdr[i] = m3dhdr[smmap[i]]     */

char zmap[30]=  { 0x63, 0x62, 0x58, 0x59, 0x60, 0x61, 0x4a, 0x4b,
				  0x64, 0x65, 0x5f, 0x5d, 0x00, 0x5a, 0x5b, 0x50,
				  0x51, 0x52, 0x53, 0x54, 0x55, 0x57, 0x56, 0x4c,
				  0x4d, 0x4e, 0x4f, 0x5e, 0x5e, 0x5e };
				  /* M3 to .Z80 register mapping: zhdr[i] = m3dhdr[zmap[i]] */
				  /* [12] is dummy and needs fixup, with [27], [28], [29]   */


char *infile;
char *outfile;
char *progname;

void m3csna (void);
void snacm3 (void);
void m3cz80 (void);
void notm3 (char *file);
void notsna (char *file);
void wrerr (char *file);
void msword (int a, int b);
void msbyte (int a, int b);
void m3dcomp (FILE *infp, FILE *outfp, unsigned short blklen);
char *mkdname (char *inname);
void set_sound(unsigned short flag128);
void build_48Kimage(FILE *infp);
void build_16Kimage(FILE *infp, unsigned short blen);
void write_image(unsigned short page_num, unsigned short byte1, FILE *outfp);
void m3dcomp_image(FILE *infp, unsigned short blkstart, unsigned short blklen);
void m3dcomp2_image(FILE *infp, unsigned short blkstart, unsigned short blklen);

/* ------------------------------------------------------------------------- */
int main (int argc, char *argv[])
  {
  short n;

  if (argc < 2)
	{
	fprintf(stderr,"%s: Invalid usage. Specify options, source and destination.\n",AV0);
	exit(1);
	}
  if (argv[1][0]!='-')
    {
    fprintf(stderr,"%s: Invalid syntax. Type \"%s -h\" for help.\n",AV0,AV0);
    exit(1);
    }

  if ((argv[1][1]=='h') || (argv[1][1]=='H') || (argv[1][1]=='?'))
    {
	fprintf(stderr,"%s v2.00 - J. Elliott, 5 August 1995.\n",AV0);
	fprintf(stderr,"              Ian Cull, 28 January 1996.\n\n");
	fprintf(stderr,"Syntax: %s -ms m3gamec game.sna {-Bn}\n",AV0);
	fprintf(stderr,"        %s -sm game.sna m3game  {-L}\n\n",AV0);
	fprintf(stderr,"        %s -mz m3gamec game.z80 {-Bn}\n\n",AV0);
    fprintf(stderr,"-Bn: Set border colour 0-7. Default is 0.\n");
    fprintf(stderr,"-L:  Lock resulting M3 snapshot.\n");
	return(0);
    }
  if (argc<4)
    {
    fprintf(stderr,"%s: Invalid syntax. Type \"%s -h\" for help.\n",AV0,AV0);
    exit(1);
    }

  for (n=4 ; n<argc ; n++)
    {
    if (argv[n][0]=='-')
      {
      switch (argv[n][1])
        {
        case 'L' : case 'l' : lock=0x20; break;
		case 'B' : case 'b' : if ((argv[n][2] < '0' ) || (argv[n][2] > '7'))
                                   {
                                   fprintf(stderr,"%s: Invalid colour: %s\n",AV0,argv[n][2]);
                                   exit(1);
                                   }
                              else border=(argv[n][2] & 7);
                              break;
        default: fprintf(stderr,"%s: Invalid option: %s\n",AV0,argv[n]);
                 exit(1);
        }
      }
	}

  infile = argv[2];
  outfile = argv[3];
  progname = AV0;

  if   ((argv[1][1]=='m') || (argv[1][1]=='M'))
    if ((argv[1][2]=='s') || (argv[1][2]=='S'))
      {
	  m3csna();
	  return(0);
      };
  if   ((argv[1][1]=='S') || (argv[1][1]=='s'))
    if ((argv[1][2]=='M') || (argv[1][2]=='m'))
      {
	  snacm3();
	  return(0);
      }
  if   ((argv[1][1]=='M') || (argv[1][1]=='m'))
	if ((argv[1][2]=='Z') || (argv[1][2]=='z'))
	  {
	  m3cz80();
	  return(0);
	  }
  fprintf(stderr,"%s: Invalid syntax. Type \"%s -h\" for help.\n",AV0,AV0);
  exit(1);
  }

/* ------------------------------------------------------------------------- */
void m3csna ()
{
FILE *infp,*outfp;
unsigned short b1len,b2len,bptr,n;

if ((infp=fopen(infile,"rb"))==NULL)
  {
  fprintf(stderr,"%s: Cannot open %s\n",progname,infile);
  exit(1);
  }
if ((outfp=fopen(outfile,"wb"))==NULL)
  {
  fprintf(stderr,"%s: Cannot create %s\n",progname,outfile);
  exit(1);
  }
if (fread(m3dhdr,1,0xD3,infp) < 0xD3) notm3(infile);
if (m3dhdr[32] & 4) notm3(infile);

for (n=0;n<25;n++) snahdr[n]=m3dhdr[smmap[n]];  /* Registers */
snahdr[19] &= 4;                                /* Clear other bits of iff2 */
snahdr[25] = (uchr)(1+((int)m3dhdr[0x5e] & 1)); /* interrupt mode */
snahdr[26] = border;                            /* Border */

if (fwrite(snahdr,1,27,outfp) < 27) wrerr(outfile);

/* SNA header written */

b1len=m3dhdr[0x1A]+256*m3dhdr[0x1b];
b2len=m3dhdr[0x1C]+256*m3dhdr[0x1d];

/* Transfer screen */

if (fseek(infp,(long)b1len,SEEK_SET) < 0) notm3(infile);
if (b2len==0x1B00)
     for (bptr=0x1B00;bptr>0;bptr--) fputc(fgetc(infp),outfp);
else m3dcomp(infp,outfp,b2len);

/* Transfer sysvars */

if (fseek(infp,(long)(b1len+b2len),SEEK_SET) < 0) notm3(infile);
for (bptr=0x05B3;bptr>0;bptr--) fputc(fgetc(infp),outfp);

/* Transfer remainder */

if (fseek(infp,(long)0xD3,SEEK_SET) < 0) notm3(infile);
if (b1len==0xA020)
     for (bptr=0x9F4D;bptr>0;bptr--) fputc(fgetc(infp),outfp);
else m3dcomp(infp,outfp,(b1len-0xD3));

if (fclose(outfp)) wrerr(outfile);
if (fclose(infp)) notm3(infile);

}

/* ------------------------------------------------------------------------- */
void snacm3 ()
{
FILE *infp,*loader,*data;
unsigned short n,sp;
char *dataname;

if ((infp=fopen(infile,"rb"))==NULL)
  {
  fprintf(stderr,"%s: Cannot open %s\n",progname,infile);
  exit(1);
  }
if ((loader=fopen(outfile,"wb"))==NULL)
  {
  fprintf(stderr,"%s: Cannot create %s\n",progname,outfile);
  exit(1);
  }

/* Generate datafile name & put it in the loader.*/

dataname=mkdname(outfile);

/* Data filename developed. Same as loader, but last component has at most 8
  characters and ends with a "c". */

if ((data=fopen(dataname,"wb"))==NULL)
  {
  fprintf(stderr,"%s: Cannot create %s\n",progname,outfile);
  exit(1);
  }

m3loader[0xF2]=0x20;
m3loader[0xF3]=0xA0;   /* Block length=0x */
m3loader[0x102]=0x71;
m3loader[0x103]=0x00;  /* disc file */

if (fwrite(m3loader,1,0x169,loader) < 0x169) wrerr(outfile);
if (fclose(loader)) wrerr(outfile);

/* M3 loader done. */

if (fread(snahdr,1,27,infp) < 27) notsna(infile);

for (n=0;n<25;n++) m3dhdr[smmap[n]]=snahdr[n];  /* Registers */
m3dhdr[0x5c] = (uchr) (0x10 | lock);            /* Mb */
m3dhdr[0x5e] &= 0x04;
m3dhdr[0x5e] |= 0x40;                           /* Mi contains iff2 & 1ffd */
if (snahdr[25]==2) m3dhdr[0x5e] |= 0x01;        /* Mi contains all data */

m3dhdr[0]=0; /* Not compressed */

m3dhdr[0x16]=0xB3;
m3dhdr[0x17]=0x60;  /* Load address of data block 1 = 0x60B3 */
m3dhdr[0x1A]=0x20;
m3dhdr[0x1B]=0xA0;  /* Length of data block 1 = 0xA020 */
m3dhdr[0x1C]=0x00;
m3dhdr[0x1D]=0x1B;  /* Length of data block 2 = 0x1B00 */
m3dhdr[0x1E]=0x00;
m3dhdr[0x1F]=0x00;  /* Block 1 not compressed */
m3dhdr[0x20]=0x00;  /* 48k snapshot to disc */
sp=snahdr[0x17]+256*snahdr[0x18];
if (sp<0x4000) fprintf(stderr,"Warning: Stack pointer is in the ROM.\n");
   else {
		if (fseek(infp,(long)(sp-0x3FE5),SEEK_SET) < 0) notsna(infile);
		m3dhdr[0x4A]=(uchr)fgetc(infp);
		m3dhdr[0x4B]=(uchr)fgetc(infp);  /* Get PC off the stack */
        }
if (fseek(infp,(long)0x20CE,SEEK_SET) < 0) notsna(infile);

if (fwrite(m3dhdr,1,0xD3,data) < 0xD3) wrerr(dataname);

/* Header written */

/* Transfer main block */

if (fseek(infp,(long)0x20CE,SEEK_SET) < 0) notsna(infile);
for (n=0x9F4D;n>0;n--) fputc(fgetc(infp),data);

/* Transfer screen */

if (fseek(infp,(long)0x1B,SEEK_SET) < 0) notsna(infile);
for (n=0x1B00;n>0;n--) fputc(fgetc(infp),data);

/* Transfer sysvars */

if (fseek(infp,(long)0x1B1B,SEEK_SET) < 0) notsna(infile);
for (n=0x05B3;n>0;n--) fputc(fgetc(infp),data);

if (fclose(data)) wrerr(dataname);
if (fclose(infp)) notsna(infile);
free(dataname);
}

/* ------------------------------------------------------------------------- */
void m3cz80 ()
{
FILE *infp,*outfp;
unsigned short n,flag128,blen;

if ((infp=fopen(infile,"rb"))==NULL)
  {
  fprintf(stderr,"%s: Cannot open %s\n",progname,infile);
  exit(1);
  }
if ((outfp=fopen(outfile,"wb"))==NULL)
  {
  fprintf(stderr,"%s: Cannot create %s\n",progname,outfile);
  exit(1);
  }
if (fread(m3dhdr,1,0xD3,infp) < 0xD3) notm3(infile);

flag128 = m3dhdr[32] & 4;

/* Build V1.45 Z80 header */
for (n=0;n<30;n++) zhdr[n]=m3dhdr[zmap[n]];     /* Registers */
zhdr[12] = (border<<1) + ((zhdr[11]&0x80)>>7);  /* Border & R msb (uncompressed) */
zhdr[12] &= 0x0F & 0x20;                        /* flag compressed file */
zhdr[27] &= 0x04;                               /* DI if zero */
zhdr[28] &= 0x04;                               /* ??? */
zhdr[29] = (zhdr[29]&1) + 1;                    /* interrupt mode */

/* Increment SP by two - remove PC from top of stack */
if (zhdr[8]>253) zhdr[9]++;
zhdr[8]++;
zhdr[8]++;

/* Convert to V2.01 Z80 header */
zhdr[30] = 23; zhdr[31] = 0;                    /* Extended header length */
zhdr[32] = zhdr[6]; zhdr[33] = zhdr[7];         /* Copy PC */
zhdr[6] = 0; zhdr[7] = 0;                       /* Zero old PC */
zhdr[34] = ((m3dhdr[32]&0x04)>>2)*3;            /* 48K/128K */
zhdr[35] = m3dhdr[0x5c];                        /* Port 7FFDh */
zhdr[36] = 0;                                   /* IF1 not paged in */
zhdr[37] = 0;                                   /* Emulations off */

/* Set sound chip registers */
set_sound(flag128);

/* Write Z80 header */
if (fwrite(zhdr,1,55,outfp) < 55) wrerr(outfile);


/* Convert three M3 blocks to a single 48K image */
build_48Kimage(infp);

/* Write three 16K blocks to Z80 file */
if (flag128)
{
	write_image(8,0x0000,outfp);                /* page 8 - 4000-7FFF */
	write_image(5,0x4000,outfp);                /* page 5 - 8000-BFFF */
	write_image(3,0x8000,outfp);                /* page 3 - C000-FFFF */
}
else
{
	write_image(8,0x0000,outfp);                /* page 8 - 4000-7FFF */
	write_image(4,0x4000,outfp);                /* page 4 - 8000-BFFF */
	write_image(5,0x8000,outfp);                /* page 5 - C000-FFFF */
};

/* Write 128K blocks if present */
if (flag128)
{
	if (m3dhdr[0x38]&0x01)
	{
		blen=m3dhdr[0x3A]+256*m3dhdr[0x3B];
		build_16Kimage(infp, blen);
		write_image(4,0x0000,outfp);                /* page 4 - page 1 */
	}
	if (m3dhdr[0x38]&0x02)
	{
		blen=m3dhdr[0x3C]+256*m3dhdr[0x3D];
		build_16Kimage(infp, blen);
		write_image(6,0x0000,outfp);                /* page 6 - page 3 */
	}
	if (m3dhdr[0x38]&0x04)
	{
		blen=m3dhdr[0x3E]+256*m3dhdr[0x3F];
		build_16Kimage(infp, blen);
		write_image(7,0x0000,outfp);                /* page 7 - page 4 */
	}
	if (m3dhdr[0x38]&0x08)
	{
		blen=m3dhdr[0x40]+256*m3dhdr[0x41];
		build_16Kimage(infp, blen);
		write_image(9,0x0000,outfp);                /* page 9 - page 6 */
	}
	if (m3dhdr[0x39]&0x08)
	{
		blen=m3dhdr[0x42]+256*m3dhdr[0x43];
		build_16Kimage(infp, blen);
		write_image(10,0x0000,outfp);                /* page 10 - page 7 */
	}
}

if (fclose(outfp)) wrerr(outfile);
if (fclose(infp)) notm3(infile);

}

/* ------------------------------------------------------------------------- */
void set_sound(unsigned short flag128)
{
unsigned short n,warnflag;
char *loadername;
FILE *loaderp;
uchr loaderbuf[0x115];

/* Initially zero all sound chip registers */
for (n=38;n<55;n++) zhdr[n] = 0;

/* If 128K snapshot, try to access M3 loader file and get sound chip values */
if (flag128)
	{
	warnflag = 0;

	loadername=malloc(strlen(infile));      /* Make a copy of M3 filename */
	if (loadername==NULL)
		{
		fprintf(stderr,"Out of memory\n");
		exit(1);
		}
	strcpy(loadername,infile);

	n=strlen(loadername);                   /* Point to last character of filename */
	n--;

	if (loadername[n]=='c' || loadername[n]=='C')
		loadername[n]='\0';                 /* Remove trailing C from filename */

	if ((loaderp=fopen(loadername,"rb"))==NULL)
		warnflag = 1;
	else
		{
		if (fread(loaderbuf,1,0x115,loaderp) < 0x115)
			warnflag = 2;
		else
			{
			if (loaderbuf[0x114] != 0x0D)
				warnflag = 3;
			else
				{
				for (n=0;n<16;n++) zhdr[n+39] = loaderbuf[n+0x104];
				}
			}
		}

	if (warnflag!=0)
		printf("128K snapshot sound registers not initialised from loader file\n");

	free(loadername);
	}
}

/* ------------------------------------------------------------------------- */
void build_48Kimage(FILE *infp)
{
unsigned short b1len,b2len,bptr,blkstart,n;

b1len=m3dhdr[0x1A]+256*m3dhdr[0x1b];
b2len=m3dhdr[0x1C]+256*m3dhdr[0x1d];

/* Erase block first */
for (bptr=0x0000;bptr<0xC000;bptr++) image[bptr] = 0;

/* Transfer main block to image*/
if (fseek(infp,(long)0xD3,SEEK_SET) < 0) notm3(infile);
blkstart = 0x20b3;
if (b1len==0xA020)
	 for (bptr=0x9F4D;bptr>0;bptr--) image[blkstart++] = (uchr)fgetc(infp);
else m3dcomp_image(infp,blkstart,(b1len-0xD3));

/* Transfer screen to image */
if (fseek(infp,(long)b1len,SEEK_SET) < 0) notm3(infile);
blkstart = 0x0000;
if (b2len==0x1B00)
	 for (bptr=0x1B00;bptr>0;bptr--) image[blkstart++] = (uchr)fgetc(infp);
else m3dcomp_image(infp,blkstart,b2len);

/* Transfer sysvars to image */
if (fseek(infp,(long)(b1len+b2len),SEEK_SET) < 0) notm3(infile);
blkstart = 0x1b00;
for (bptr=0x05B3;bptr>0;bptr--) image[blkstart++] = (uchr)fgetc(infp);

/* Now file pointer is just past sysvars - may be at start of 128K data */
}

/* ------------------------------------------------------------------------- */
void build_16Kimage(FILE *infp, unsigned short blen)
{
unsigned short bptr,blkstart,blksize,n;
int c;


/* Erase block first */
for (bptr=0x0000;bptr<0x4000;bptr++) image[bptr] = 0;

/* Transfer a 16K block to image - assume file at first byte of block */
blkstart = 0x0000;
if (blen==0x4000)
	 for (bptr=0x0000;bptr<0x4000;bptr++)
	 {
		c = fgetc(infp);
		if (c==EOF) notm3(infile);
		image[blkstart++] = (uchr)c;
	 }
else
	m3dcomp2_image(infp,blkstart,blen);
}

/* ------------------------------------------------------------------------- */
char *mkdname(char *inname)
{
char *dataname;
short bptr,n;

dataname=malloc(2+strlen(inname));  /* allocate space for filename */
if (dataname==NULL)
  {
  fprintf(stderr,"Out of memory\n");
  exit(1);
  }

strcpy(dataname,inname);

bptr=strlen(dataname);
bptr--;                /* last character */

for (n=0;n<1;bptr--)
  {                               /* Just get the last filename component */
  if (   dataname[bptr]=='/'      /* Unix directory separator */
      || dataname[bptr]=='\\'     /* DOS directory separator */
      || dataname[bptr]==':')     /* DOS & CP/M drive specifier */
    {
    bptr++;
    bptr++;
    n++;
    }
  if ((bptr==0) && (n==0))        /* Hit the beginning of the name? */
    {
    n++;
    bptr++;
    }
  }

/* dataname[bptr] onwards is loader filename (no directories) */

for (n=0;n<8;n++)
  {
  if (dataname[n+bptr] == '\0')
    {
    dataname[n+bptr]='c';
    dataname[++n+bptr]='\0';
    n=10;
    }
  }
if ((n!=10) && (dataname[bptr+7]))
  {
  dataname[bptr+7]='c';
  dataname[bptr+8]='\0';
  }

for (n=0;n<9;n++)
  {
  if (dataname[bptr+n]) m3loader[0xE9+n]=dataname[bptr+n]; /* not end of name */
  else m3loader[0xE9+n]=0xFF;                              /* end of name */
  }

return (dataname);
}

/* ------------------------------------------------------------------------- */
void m3dcomp(FILE *infp, FILE *outfp, unsigned short blklen)
{
uchr buf[6];
unsigned short m,o,rlelen;
char n,p;

for(n=0;n<6;n++)  buf[n]=(uchr)fgetc(infp);   /* Initialise buffer */

for(m=blklen;m>0;)
  {
  if ((buf[3]==0x37) && (buf[4]==0xED) && (buf[5]==0xCB))  /* RLE signature */
    {
    p=buf[2];
    rlelen=buf[0]+256*buf[1];

    for(o=rlelen;o>0;o--) fputc(p,outfp);  /* Un-RLE the data */
	for(n=0;n<6;n++)  buf[n]=(uchr)fgetc(infp);  /* Reinitialise buffer */
    if (m<6) m=0; else m-=6;
    }
  else
    {
    fputc(buf[0],outfp);
    for(n=0;n<5;n++)  buf[n]=buf[n+1];  /* Do 1 byte */
	buf[5]=(uchr)fgetc(infp);
    m--;
    }
  }
}

/* ------------------------------------------------------------------------- */
void m3dcomp_image(FILE *infp, unsigned short blkstart, unsigned short blklen)
{
uchr buf[6];
unsigned short m,o,rlelen;
char n,p;

for(n=0;n<6;n++)  buf[n]=(uchr)fgetc(infp);   /* Initialise buffer */

for(m=blklen;m>0;)
  {
  if ((buf[3]==0x37) && (buf[4]==0xED) && (buf[5]==0xCB))  /* RLE signature */
	{
	p=buf[2];
	rlelen=buf[0]+256*buf[1];

	for(o=rlelen;o>0;o--) image[blkstart++] = p;  /* Un-RLE the data */
	for(n=0;n<6;n++)  buf[n]=(uchr)fgetc(infp);  /* Reinitialise buffer */
	if (m<6) m=0; else m-=6;
	}
  else
	{
	image[blkstart++] = buf[0];
	for(n=0;n<5;n++)  buf[n]=buf[n+1];  /* Do 1 byte */
	buf[5]=(uchr)fgetc(infp);
	m--;
	}
  }
}

/* ------------------------------------------------------------------------- */
void m3dcomp2_image(FILE *infp, unsigned short blkstart, unsigned short blklen)
{
uchr buf[6];
unsigned short blksize,m,o,rlelen;
long filepos;
char n,p;

filepos = ftell(infp);
blksize = (uchr)fgetc(infp);
blksize = (uchr)fgetc(infp)*256 + blksize;
if (blksize!=blklen) notm3(infile);

if (fseek(infp,filepos+(long)blksize-2,SEEK_SET) < 0) notm3(infile);
for(n=0;n<2;n++)  buf[n]=(uchr)fgetc(infp);   /* Initialise first two bytes of buffer */

if (fseek(infp,filepos+2,SEEK_SET) < 0) notm3(infile);
for(n=2;n<6;n++)  buf[n]=(uchr)fgetc(infp);   /* Initialise rest of buffer */

for(m=blklen;m>0;)
  {
  if ((buf[3]==0x37) && (buf[4]==0xED) && (buf[5]==0xCB))  /* RLE signature */
	{
	p=buf[2];
	rlelen=buf[0]+256*buf[1];

	for(o=rlelen;o>0;o--) image[blkstart++] = p;  /* Un-RLE the data */
	for(n=0;n<6;n++)  buf[n]=(uchr)fgetc(infp);  /* Reinitialise buffer */
	if (m<6) m=0; else m-=6;
	}
  else
	{
	image[blkstart++] = buf[0];
	for(n=0;n<5;n++)  buf[n]=buf[n+1];  /* Do 1 byte */
	buf[5]=(uchr)fgetc(infp);
	m--;
	}
  }

if (fseek(infp,filepos+(long)blksize,SEEK_SET) < 0) notm3(infile);

}

/* ------------------------------------------------------------------------- */
void notm3 (char *file)
{
  fprintf(stderr,"Read error or not a 48k M3 datafile: %s\n",file);
  exit(1);
}

void notsna (char *file)
{
  fprintf(stderr,"Read error or not a .SNA file: %s\n",file);
  exit(1);
}

void wrerr (char *file)
{
  fprintf(stderr,"Write error: %s\n",file);
  exit(1);
}

/* ------------------------------------------------------------------------- */
void write_image(unsigned short page_num, unsigned short byte1, FILE *outfp)
{
unsigned short n,length,tel;
unsigned char byte;
unsigned char *source, *dest;

/* Compress the 16K block of memory (on to itself) */
length = 0x4000;
source = image+byte1;
dest = image2;
tel = 0;
	do {
		if (tel==0)
		{
			byte=*source++;
			tel=1;
			length--;
		} else
		if ((tel<255)&&(byte==*source)&&(length!=0))
		{
			source++;
			tel++;
			length--;
		} else
		if ((tel>=5)||((byte==0xED)&&(tel>=2)))
		{
			*dest++ = 0xED;
			*dest++ = 0xED;
			*dest++ = (uchr)tel;
			*dest++ = byte;
			tel=0;
		} else
		{
			for (n=0;n<tel;n++) *dest++ = byte;
			tel=0;
			if ((byte==0xED)&&(length!=0))
			{
				*dest++ = *source++;
				length--;
			}
		}
	} while ((length!=0)||(tel!=0));

/* Calculate new length of compressed block */
length = dest-image2;

/* Write block header information */
fputc((uchr)(length & 0xFF),outfp);             /* length */
fputc((uchr)(length >> 8),outfp);
fputc((uchr)page_num,outfp);                    /* page number */

/* Write the block */
for (n=0;n<length;n++)
	fputc(image2[n],outfp);

}
